# Data Description 

Contains a MAT file called trainedDLCHOMP.mat that was created using the dlCHOMP feature of the Robotics System Toolbox in MATLAB® R2024a.
Directory structure can be seen below.

```
parentFolder
│   README.md
│	license.txt
│   trainedDLCHOMP.mat    
```

The trainedDLCHOMP.mat file contains a pretrained dlCHOMP object created for a universalUR5e robot to output a trajectory with 40 waypoints.

This object was trained on a dataset of 5000 data samples generated using the generateSamples() function on the same dlCHOMP object.

The random number generator algorithm was set to "twister" and its seed set to 100 before the generateSamples() call that generated the data samples. This ensured exact replication of the data samples.

Data augmentation was performed by symmetrically flipping the data sample trajectories thereby adding 5000 more data samples to the dataset.

In this manner the final dataset had 10000 data samples.

The dlCHOMP object was then trained using the trainDLCHOMP() function.

The MAT file trainedDLCHOMP.mat contains:
	- robotName: The name of the robot for which the data was generated, in this case it is "universalUR5e".
	- dataOptions: The dlCHOMPDataOptions object that was used to generate the training data used to train the dlCHOMP object.
	- trainOptions: The trainingOptions object that was used to train the dlCHOMP object.
	- trainLossType: The loss type that was used to train the dlCHOMP object.
	- trainedDLCHOMP: The trained dlCHOMP object.
	- trainedNetwork: The trained dlnetwork object. This is the same as the value of the Network property of trainedDLCHOMP. 
	- trainInfo: The training progress plot that was obtained while training the dlCHOMP object.
	- unseenObstacles: An unseen environment of spherical obstacles for testing.
	- unseenStart: An unseen start configuration for testing.
	- unseenGoal:  An unseen goal configuration for testing.
	
The dlCHOMP object was trained and saved by running the following code snippet in MATLAB R2024a:
```
% Load robot of interest
robotName = "universalUR5e";
rbt = loadrobot(robotName,DataFormat="row");

% Obtain capsule approximation
rbtcap = capsuleApproximation(rbt);

% Update pose and geometry of upper arm link capsule
% Note: This is needed to ensure a more accurate
% collision geometry which in turn reduces self-collisions,
% leading to faster training data generation times.
capsule = getCapsules(rbtcap,"upper_arm_link");
oldPose = capsule{:}.Pose;
newPose = oldPose;
newPose(3,4) = 0.15;
updatePose(rbtcap,"upper_arm_link",newPose);
updateGeometry(rbtcap,"upper_arm_link",[0.1 0.4]);

% Update joint position limits
% Note: This is needed to ensure 1-to-1 mapping between
% configuration in the 3D space to angles in joint space.
% This helps ensure the training loss can converge leading to
% a more accurate trained network.
rbtAdjusted = rbtcap.RigidBodyTree;
revoluteJointBodyNames = ["shoulder_link";
                 "upper_arm_link";
                 "forearm_link";
                 "wrist_1_link";
                 "wrist_2_link";
                 "wrist_3_link" ];
for jointBodyIdx=1:numel(revoluteJointBodyNames)
    currJointBodyName = revoluteJointBodyNames(jointBodyIdx);
    currBody = getBody(rbtAdjusted,currJointBodyName);
    currBody.Joint.PositionLimits = [-pi pi];
end
rbt = rbtAdjusted;

% Create basis point set encoder
% Set seed to exactly replicate basis point set
seed = 100;
rng(seed,"twister");
encoder = bpsEncoder("uniform-ball-3d",1e4);

% Set number of waypoints
numWaypoints = 40;

% Create dlCHOMP object
dlchomp = dlCHOMP(rbt,encoder,numWaypoints);
dlchomp.SolverOptions.LearningRate = 2;
dlchomp.SolverOptions.FunctionTolerance=1e-4;
dlchomp.SolverOptions.MaxIterations = 200;
dlchomp.SmoothnessOptions.SmoothnessCostWeight=0.001;
dlchomp.CollisionOptions.CollisionCostWeight=120;
dlchomp.CollisionOptions.IgnoreSelfCollision = false;

% Generate training and validation datasets having 4000 and 1000 samples respectively
% Set seed to exactly replicate the set of data samples.
rng(seed,"twister");
dataOptions = dlCHOMPDataOptions(...
	RadiusRangeOfObstacles=[0.05 0.20],...
	CountRangeOfObstacles=[20 25],...
	MinDistanceFromBase=0.21,...
	NumSamples=5000,...
	ValidationSplit=0.2);
dataFolderName = "./allData";
[trainDS,validDS] = generateSamples(dlchomp,dataOptions,dataFolderName);

% Create an new transformed dataset by flipping the training dataset samples
trainDSFlipped = transform(trainDS,...
					@(x) { ...
					flipWaypoints(x{1},numel(x{1})/2),...
					x{2},...
					flipWaypoints(x{3},numel(x{1})/2),...
					});
					
% Create a new transformed dataset by flipping the validation dataset samples
validDSFlipped = transform(validDS,...
					@(x) { ...
					flipWaypoints(x{1},numel(x{1})/2),...
					x{2},...
					flipWaypoints(x{3},numel(x{1})/2),...
					});

% Combine the original training dataset and flipped training dataset
% to create one large augmented training dataset of 8000 samples
combinedTrainDS = combine(trainDS,trainDSFlipped,ReadOrder="sequential");

% Combine the original validation dataset and flipped validation dataset
% to create one large augmented validation dataset of 2000 samples
combinedValidDS = combine(validDS,validDSFlipped,ReadOrder="sequential");

% Create training options object
trainOptions = trainingOptions(
	"adam", ...
	"MaxEpochs",300, ...
	"MiniBatchSize",256, ...
	"InitialLearnRate",0.001, ...
	"LearnRateDropFactor",0.9, ...
	"LearnRateDropPeriod",10, ...
	"LearnRateSchedule","piecewise", ...
	"ValidationData",combinedValidDS, ...
	"ExecutionEnvironment","parallel", ...
	"Plots","training-progress", ...
	'Shuffle','every-epoch');

% Train dlCHOMP
trainLossType = "mean-squared-error";
trainInfo = trainDLCHOMP(dlchomp,combinedTrainDS,trainLossType,trainOptions);
trainedDLCHOMP = dlchomp;
trainedNetwork = dlchomp.Network;

% Save trained dlCHOMP
save("trainedDLCHOMP.mat",...
	"robotName","trainedDLCHOMP","trainedNetwork","trainInfo",...
	"dataOptions","trainOptions","trainLossType");
	
% Helper function
function flippedX = flipWaypoints(x,jointSize)
%flipWaypoints Flips a column vector x in a part-wise fashion
%   Flips a column vector x by dividing its elements into bins of size
%   jointSize in a sequential fashion and then reversing the ordering of
%   such bins. Waypoint is synonymous with a bin here.
%
%   Example 1:
%       xBefore = 1:10
%       xAfter = flipWaypoints(xBefore,2)
%
%   Example 2:
%       flippedTrainDS = transform(trainDS, ...
%           @(x) { ...
%           flipWaypoints(x{1},numel(x{1})/2),...
%           x{2},...
%           flipWaypoints(x{3},numel(x{1})/2),...
%       });
    y = reshape(x,jointSize,[])';
    z = flip(y);
    flippedX = reshape(z',[],1)';
end
```

Copyright 2024 The MathWorks, Inc.